import asyncio


class Throttle:
    """
    Helper class that wrap a function with a given max execution rate.
    By default the rate is set to execute no more than once a second.

    :param fn: the function to call.
    :type fn: function

    :param ts: Number of seconds to wait before the next execution.
    :type ts: float
    """

    def __init__(self, fn, ts=1):
        self._ts = ts
        self._fn = fn
        self._requests = 0
        self._pending = False
        self._pending_task = None
        self._last_args = []
        self._last_kwargs = {}

    @property
    def rate(self):
        """Number of maximum executions per second"""
        return 1.0 / self._ts

    @rate.setter
    def rate(self, rate):
        """Update the maximum number of executions per seconds"""
        self._ts = 1.0 / rate

    @property
    def delta_t(self):
        """Number of seconds to wait between execution"""
        return self._ts

    @delta_t.setter
    def delta_t(self, seconds):
        """Update the number of seconds to wait between execution"""
        self._ts = seconds

    async def _trottle(self):
        self._pending = True
        if self._requests:
            self._fn(*self._last_args, **self._last_kwargs)
            self._requests = 0

        await asyncio.sleep(self._ts)
        if self._requests > 0:
            await self._trottle()
        self._pending = False

    def __call__(self, *args, **kwargs):
        """Function call wrapper that will throttle the actual function provided at construction"""
        self._requests += 1
        self._last_args = args
        self._last_kwargs = kwargs

        if not self._pending:
            self._pending_task = asyncio.create_task(self._trottle())



import asyncio
import logging

__all__ = [
    "create_task",
    "decorate_task",
    "get_number_of_pending_tasks",
    "handle_task_result",
    "task",
]

PENDING_TASKS = set()


def get_number_of_pending_tasks():
    return len(PENDING_TASKS)


def handle_task_result(task: asyncio.Task) -> None:
    try:
        task.result()
    except asyncio.CancelledError:
        pass  # Task cancellation should not be logged as an error.
    except Exception:  # pylint: disable=broad-except
        logging.exception("Exception raised by task = %r", task)


def decorate_task(task):
    """
    Decorate a task by attaching a done callback so any exception or error could
    be caught and reported.

    :param task: A coroutine to execute as an independent task
    :type task: asyncio.Task

    :return: The same task object
    :rtype: asyncio.Task
    """
    PENDING_TASKS.add(task)
    task.add_done_callback(handle_task_result)
    task.add_done_callback(PENDING_TASKS.discard)
    return task


def create_task(coroutine, loop=None):
    """
    Create a task from a coroutine while also attaching a done callback so any
    exception or error could be caught and reported.

    :param coroutine: A coroutine to execute as an independent task
    :param loop: Optionally provide the loop on which the task should be
                 scheduled on. By default we will use the current running loop.

    :return: The decorated task
    :rtype: asyncio.Task
    """
    if loop is None:
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

    return decorate_task(loop.create_task(coroutine))


def task(func):
    """Function decorator to make its async execution within a task"""

    def wrapper(*args, **kwargs):
        create_task(func(*args, **kwargs))

    return wrapper



import functools
import inspect
import logging

logger = logging.getLogger(__name__)

# logging.basicConfig(level=logging.DEBUG)


def can_be_decorated(x):
    return inspect.ismethod(x) or inspect.isfunction(x)


class TrameApp:
    """
    Class Decorator for trame application.
    This decorator can be used to automatically register methods to
    state.change, controller, trigger and life_cycle.
    The parameter are used to define where to fine the server instance
    within the object and where a namespace for the variable name is stored.


    .. code-block:: python

        @TrameApp()
        class ExampleApp:
            def __init__(self):
                self.server = get_server()

            @change("var_name_1", "var_name_n")
            def on_state_change(**kwargs):
                pass

            @controller.set("hello")
            def hello(**kwargs):
                pass

    """

    def __init__(self, server="server", namespace=""):
        self.server_name = server
        self.namespace_prefix = namespace

    def __call__(self, klass):
        self.klass_init = klass.__init__

        @functools.wraps(klass.__init__, updated=())
        def wrapped_init(instance, *args, **kwargs):
            logger.debug("Create instance")
            self.klass_init(instance, *args, **kwargs)

            logger.debug("Instance created")

            server = getattr(instance, self.server_name)
            prefix = (
                getattr(instance, self.namespace_prefix)
                if self.namespace_prefix
                else ""
            )

            logger.debug("server=%s prefix=%s", server, prefix)

            # Look for method decorator
            for k in inspect.getmembers(instance.__class__, can_be_decorated):
                fn = getattr(instance, k[0])

                # Handle @state.change
                if "_trame_state_change" in fn.__dict__:
                    state_change_names = fn.__dict__["_trame_state_change"]
                    key_names = [f"{prefix}{v}" for v in state_change_names]
                    logger.debug("state.change(%s)(%s)", key_names, k[0])
                    server.state.change(*key_names)(fn)

                # Handle @trigger
                if "_trame_trigger_names" in fn.__dict__:
                    trigger_names = fn.__dict__["_trame_trigger_names"]
                    for trigger_name in trigger_names:
                        logger.debug("trigger(%s)(%s)", trigger_name, k[0])
                        server.trigger(f"{trigger_name}")(fn)
                        if prefix:
                            logger.debug(
                                "trigger(%s%s)(%s)", prefix, trigger_name, k[0]
                            )
                            server.trigger(f"{prefix}{trigger_name}")(fn)

                # Handle @ctrl.[add, once, add_task, set]
                if "_trame_controller" in fn.__dict__:
                    actions = fn.__dict__["_trame_controller"]
                    for action in actions:
                        name = action.get("name")
                        method = action.get("method")
                        decorate = getattr(server.controller, method)
                        logger.debug("ctrl.%s(%s)(%s)", method, name, k[0])
                        decorate(name)(fn)
                        if prefix:
                            logger.debug(
                                "ctrl.%s(%s%s)(%s)", method, prefix, name, k[0]
                            )
                            decorate(f"{prefix}{name}")(fn)

        klass.__init__ = wrapped_init
        return klass


def change(*args):
    """Method decorator for state change"""

    def decorate(f):
        if not hasattr(f, "_trame_state_change"):
            f._trame_state_change = []
        f._trame_state_change.extend(args)
        return f

    return decorate


def trigger(*args):
    """Method decorator to assign a trigger name to a function"""

    def decorate(f):
        if not hasattr(f, "_trame_trigger_names"):
            f._trame_trigger_names = []
        f._trame_trigger_names.extend(args)
        return f

    return decorate


def controller_decorator(method):
    def decorator(*args):
        def decorate(f):
            if not hasattr(f, "_trame_controller"):
                f._trame_controller = []
            for name in args:
                f._trame_controller.append({"method": method, "name": name})
            return f

        return decorate

    return decorator


class Controller:
    """Controller decorators

    .. code-block:: text

        - once
        - add
        - add_task
        - set
    """

    def __init__(self):
        self.once = controller_decorator("once")
        self.add = controller_decorator("add")
        self.add_task = controller_decorator("add_task")
        self.set = controller_decorator("set")


class LifeCycle:
    """Life Cycle decorators

    .. code-block:: text

        - server_start
        - server_bind
        - server_ready
        - client_connected
        - client_exited
        - server_exited
        - server_reload
        - error
        - exception

    For asynchronous callback, you can register them as. Note that "server_bind" is not available.

    .. code-block:: text

        - server_start_task
        - server_ready_task
        - client_connected_task
        - client_exited_task
        - server_exited_task
        - server_reload_task
        - error_task
        - exception_task

    """

    def __init__(self):
        # add
        self.server_start = controller_decorator("add")("on_server_start")
        self.server_bind = controller_decorator("add")("on_server_bind")
        self.server_ready = controller_decorator("add")("on_server_ready")
        self.client_connected = controller_decorator("add")("on_client_connected")
        self.client_exited = controller_decorator("add")("on_client_exited")
        self.server_exited = controller_decorator("add")("on_server_exited")
        self.server_reload = controller_decorator("add")("on_server_reload")
        self.error = controller_decorator("add")("on_error")
        self.exception = controller_decorator("add")("on_exception")

        # add_task
        self.server_start_task = controller_decorator("add_task")("on_server_start")
        self.server_ready_task = controller_decorator("add_task")("on_server_ready")
        self.client_connected_task = controller_decorator("add_task")(
            "on_client_connected"
        )
        self.client_exited_task = controller_decorator("add_task")("on_client_exited")
        self.server_exited_task = controller_decorator("add_task")("on_server_exited")
        self.server_reload_task = controller_decorator("add_task")("on_server_reload")
        self.error_task = controller_decorator("add_task")("on_error")
        self.exception_task = controller_decorator("add_task")("on_exception")


controller = Controller()
life_cycle = LifeCycle()


__all__ = [
    "TrameApp",
    "change",
    "controller",
    "life_cycle",
    "trigger",
]



from trame.app import get_server

from trame_common.exec.asynchronous import create_task
from trame_common.obj.component import TrameComponent


class TrameApp(TrameComponent):
    """
    Base trame class that has access to a trame server instance
    on which we provide simple accessor and method decoration capabilities.
    """

    def __init__(self, server=None, client_type="vue3", ctx_name=None, **_):
        super().__init__(get_server(server, client_type=client_type), ctx_name=ctx_name)

    async def _async_display(self):
        from IPython.display import clear_output

        await self.ui.ready
        clear_output(wait=True)
        self.ui._ipython_display_()

    def _repr_html_(self):
        create_task(self._async_display())
        return "<i>Launching trame server in the background...</i>"



import inspect
import logging

__all__ = [
    "TrameComponent",
]

logger = logging.getLogger(__name__)


def can_be_decorated(x):
    return inspect.ismethod(x) or inspect.isfunction(x)


class TrameComponent:
    """
    Base trame class that has access to a trame server instance
    on which we provide simple accessor and method decoration capabilities.
    """

    def __init__(self, server, ctx_name=None, **_):
        """
        Initialize TrameComponent with its server.

        Keyword arguments:
        server -- the server to link to (default None)
        ctx_name -- name to use to bind current instance to server.context (default None)
        """
        self._server = server

        if ctx_name:
            self.ctx[ctx_name] = self

        self._bind_annotated_methods()

    @property
    def server(self):
        """Return the associated trame server instance"""
        return self._server

    @property
    def state(self):
        """Return the associated server state"""
        return self.server.state

    @property
    def ctrl(self):
        """Return the associated server controller"""
        return self.server.controller

    @property
    def ctx(self):
        """Return the associated server context"""
        return self.server.context

    def _bind_annotated_methods(self):
        # Look for method decorator
        for k in inspect.getmembers(self.__class__, can_be_decorated):
            fn = getattr(self, k[0])

            # Handle @state.change
            s_translator = self.state.translator
            if "_trame_state_change" in fn.__dict__:
                state_change_names = fn.__dict__["_trame_state_change"]
                logger.debug(
                    "state.change(%s)(%s)",
                    [f"{s_translator.translate_key(v)}" for v in state_change_names],
                    k[0],
                )
                self.state.change(*[f"{v}" for v in state_change_names])(fn)

            # Handle @trigger
            if "_trame_trigger_names" in fn.__dict__:
                trigger_names = fn.__dict__["_trame_trigger_names"]
                for trigger_name in trigger_names:
                    logger.debug("trigger(%s)(%s)", trigger_name, k[0])
                    self.server.trigger(f"{trigger_name}")(fn)

            # Handle @ctrl.[add, once, add_task, set]
            if "_trame_controller" in fn.__dict__:
                actions = fn.__dict__["_trame_controller"]
                for action in actions:
                    name = action.get("name")
                    method = action.get("method")
                    decorate = getattr(self.ctrl, method)
                    logger.debug("ctrl.%s(%s)(%s)", method, name, k[0])
                    decorate(name)(fn)

    def _unbind_annotated_methods(self):
        # Look for method decorator
        for k in inspect.getmembers(self.__class__, can_be_decorated):
            fn = getattr(self, k[0])

            # Handle @state.change
            methods_to_detach = {}
            if "_trame_state_change" in fn.__dict__:
                methods_to_detach.add(fn)

            if methods_to_detach:
                for fn_list in self.state._change_callbacks.values():
                    to_remove = set(fn_list) | methods_to_detach
                    for fn in to_remove:
                        fn_list.remove(fn)

            # Handle @trigger
            # TODO

            # Handle @ctrl
            # TODO



import sys
import time
from contextlib import contextmanager

__all__ = [
    "LOGGER",
    "Logger",
    "Timer",
    "clear_filter",
    "enable",
    "exclude",
    "include",
    "timer",
]

KEY_SIZE = 2


class Filter:
    def __init__(self):
        self._include = {}
        self._exclude = {}

    def include(self, head):
        if len(head) < KEY_SIZE:
            return

        key = head[:KEY_SIZE]
        self._include.setdefault(key, set()).add(head)

    def exclude(self, head):
        if len(head) < KEY_SIZE:
            return

        key = head[:KEY_SIZE]
        self._exclude.setdefault(key, set()).add(head)

    def clear(self, include=True, exclude=True):
        if include:
            self._include.clear()
        if exclude:
            self._exclude.clear()

    def keep(self, msg):
        if not self._exclude or not self._include or len(msg) < KEY_SIZE:
            return True

        # compute key for fast lookup
        key = msg[:KEY_SIZE]

        # Exclude
        if self._exclude:
            exclude_set = self._exclude.get(key, [])
            for head in exclude_set:
                if msg.startswith(head):
                    return False

        # Include all if empty
        if not self._include:
            return True

        # Only show includes
        if self._include:
            include_set = self._include.get(key, [])
            for head in include_set:
                if msg.startswith(head):
                    return True

        return False


class Logger:
    __slots__ = ("abort", "enabled", "filter", "log_perf", "log_perf_fps")

    def __init__(self):
        self.filter = Filter()
        self.abort = False
        self.enabled = False
        self.log_perf = None
        self.log_perf_fps = None
        self.use_print()

    def action(self, msg, duration=0):
        if self.enabled and self.filter.keep(msg):
            self.log_perf(time.perf_counter(), msg, duration)

    def use_print(self, file=sys.stderr):
        def log_perf(ts, msg, dt_ms):
            print(f"{msg:<60} {ts:10.3f} {dt_ms:8.2f} ms", file=file, flush=True)

        def log_perf_fps(ts, msg, dt_ms):
            print(
                f"{msg:<60} {ts:10.3f} {dt_ms:8.2f} ms {1000 / dt_ms:8.1f} fps",
                file=file,
                flush=True,
            )

        self.log_perf = log_perf
        self.log_perf_fps = log_perf_fps
        return self

    def use_loguru(self):
        from loguru import logger

        logger.add(
            sys.stdout,
            format="[PERF] {message}",
            level="TRACE",
            filter="trame_lumo_view.perf",
        )

        def log_perf(ts, msg, dt_ms):
            logger.trace("{:<60} {:10.3f} {:8.2f} ms", msg, ts, dt_ms)

        def log_perf_fps(ts, msg, dt_ms):
            logger.trace(
                "{:<60} {:8.3f} {:10.2f} ms {:8.1f} fps", msg, ts, dt_ms, 1000 / dt_ms
            )

        self.log_perf = log_perf
        self.log_perf_fps = log_perf_fps
        return self


class Timer:
    __slots__ = ("abort", "dt", "log", "msg", "t0")

    def __init__(self, msg: str, show_fps: bool = False):
        self.msg = msg
        self.abort = False
        self.t0 = 0
        self.dt = 0
        self.log = LOGGER.log_perf_fps if show_fps else LOGGER.log_perf

    def __enter__(self):
        self.on_start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.on_end()

    def on_start(self, *_, **__):
        self.t0 = time.perf_counter()

    def on_end(self, *_, **__):
        self.dt = (time.perf_counter() - self.t0) * 1000.0
        if self.abort:
            self.abort = False
        elif LOGGER.enabled and LOGGER.filter.keep(self.msg):
            self.log(self.t0, self.msg, self.dt)


LOGGER = Logger()


def enable(value: bool = True) -> Logger:
    """Turn performance instrumentation on or off globally."""
    LOGGER.enabled = bool(value)
    return LOGGER


def include(value: str = "") -> Logger:
    """Register filter to only include messages starting with value."""
    LOGGER.filter.include(value)
    return LOGGER


def exclude(value: str = "") -> Logger:
    """Register filter to exclude messages starting with value."""
    LOGGER.filter.exclude(value)
    return LOGGER


def clear_filter(include: bool = True, exclude: bool = True) -> Logger:
    """Clear filter."""
    LOGGER.filter.clear(include, exclude)
    return LOGGER


@contextmanager
def timer(label: str, show_fps: bool = False):
    # Skip
    if not LOGGER.enabled:
        yield LOGGER
        return

    # Time it
    t0 = time.perf_counter()
    try:
        yield LOGGER
    finally:
        if LOGGER.abort:
            LOGGER.abort = False
        else:
            dt_ms = (time.perf_counter() - t0) * 1000.0
            if show_fps:
                LOGGER.log_perf_fps(t0, label, dt_ms)
            else:
                LOGGER.log_perf(t0, label, dt_ms)



from trame_client.widgets.core import AbstractElement

__all__ = [
    "create_class",
]


def create_class(
    class_name,
    component_name,
    properties=None,
    events=None,
    module=None,
):
    """Helper for creating Widget class

    Args:
        class_name (string): name of the Python generated class
        component_name (string): name of the vue component
        properties (list, optional): List of properties mapping. Defaults to [].
        events (list, optional): List of events mapping. Defaults to [].
        module (dict, optional): Module to enable when using the class. Defaults to None.
    """
    if properties is None:
        properties = []

    if events is None:
        events = []

    def constructor(self, **kwargs):
        AbstractElement.__init__(self, component_name, **kwargs)
        if module is not None:
            self.server.enable_module(module)

        self._attr_names += properties
        self._event_names += events

    return type(
        class_name,
        (AbstractElement,),
        {"__init__": constructor},
    )
